module Ml_z_init = struct
  (* Not necessary to test.  The marshalling tests cover initialization *)
end

module Ml_z_mul_overflows = struct
  (* This function isn't exposed, and it's covered by being
     used in other tests *)
end
